/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#pragma ident	"@(#)chkey.c	1.17	94/04/25 SMI"

/*
 * Copyright (C) 1986, Sun Microsystems, Inc.
 */

/*
 * Command to change one's public key in the public key database
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <rpcsvc/ypclnt.h>
#include <pwd.h>
#include <shadow.h>
#include <crypt.h>
#include <string.h>
#include <rpcsvc/nis.h>

#include "keyclient.h"
#include <rpc/keylib.h>

#define	PK_FILES	1
#define	PK_YP		2
#define	PK_NISPLUS	3

char	program_name[256];
extern	int optind;
extern	char *optarg;
static int getpasswd(uid_t uid, char *netname, int force, char *newpass, char *secret);
static int setpublicmap(char *netname, char *public, char *secret, int database);

static void usage(void) __attribute__((noreturn));


static char PKMAP[] = "publickey.byname";
static char PKFILE[] = "/etc/publickey";
#define	MAXHOSTNAMELEN	256

int	reencrypt_only = 0;

int
main(int argc, char *argv[])
{
	unsigned keybytes;
	unsigned keybits;
	unsigned hexkeybytes;

	char	netname[MAXNETNAMELEN+1];
	char	domain[MAXNETNAMELEN+1];
	char	public[MAX_HEXKEYBYTES + 1];
	char	secret[MAX_HEXKEYBYTES + 1];
	char	crypt1[MAX_HEXKEYBYTES + KEYCHECKSUMSIZE + 1];
	char	newpass[9]; /* as per getpass() man page */
	int	status;
	uid_t	uid;
	int	force = 0;
	int	pk_database = 0;
	char	*pk_service = NULL;
	char	c;

	strcpy(program_name, argv[0]);
	while ((c = getopt(argc, argv, "fps:")) != -1) {
		switch (c) {
		case 'f':
			/*
			 * temporarily supported,
			 * not documented as of s1093.
			 */
			force = 1;
			break;
		case 'p':
			reencrypt_only = 1;
			break;
		case 's':
			if (pk_service == NULL)
				pk_service = optarg;
			else
				usage();
			break;
		default:
			usage();
		}
	}

	if (optind < argc)
		usage();

	if ((pk_database = get_pk_source(pk_service)) == 0)
		usage();

	getdomainname(domain, MAXHOSTNAMELEN);
	/*
	 * note: we're using getuid() and not geteuid() as
	 * chkey is a root set uid program.
	 */
	if (getnetnameof(netname, uid = getuid(), domain) < 0) {
		(void) fprintf(stderr,
			"%s: cannot generate netname for uid %d\n",
				program_name, uid);
		exit(1);
	}

	if (reencrypt_only)
		(void) fprintf(stdout,
			"Reencrypting key for '%s'.\n", netname);
	else
		(void) fprintf(stdout,
			"Generating new key for '%s'.\n", netname);

	setmodulus_and_root(DEFAULT_HEXMODULUS, DEFAULT_PROOT, &keybytes);
	hexkeybytes = keybytes << 1;
	keybits = keybytes << 3;

	if (! getpasswd(uid, netname, force, newpass, secret))
		exit(1);
	/* at this point geteuid() == uid */

	if (reencrypt_only) {
		char * ppub;
		if (! getpublickey_long(netname, &ppub) || !ppub) {
			(void) fprintf(stderr,
				"%s: cannot get public key for %s.\n",
				program_name, netname);
			exit(1);
		}
		strcpy(public, ppub);
		free(ppub);
	} else
		genkeys(public, secret, newpass, keybits);

	memcpy(crypt1, secret, hexkeybytes);
	memcpy(crypt1 + hexkeybytes, secret, KEYCHECKSUMSIZE);
	crypt1[hexkeybytes + KEYCHECKSUMSIZE] = 0;
#ifndef SECURE_PUBLICKEY
	xencrypt(crypt1, newpass);
#else
	xxencrypt(crypt1, newpass, netname);
#endif

	status = setpublicmap(netname, public, crypt1, pk_database);

	if (status) {
		switch (pk_database) {
		case PK_YP:
			(void) fprintf(stderr,
				"%s: unable to update NIS database (%u): %s\n",
				program_name, status, yperr_string(status));
			break;
		case PK_FILES:
			(void) fprintf(stderr,
			"%s: hence, unable to update publickey database\n",
				program_name);
			break;
		case PK_NISPLUS:
			(void) fprintf(stderr,
				"%s: unable to update nisplus database\n",
				program_name);
			break;
		default:
			(void) fprintf(stderr,
				"could not update; database %d unknown\n",
				pk_database);
		}
		exit(1);
	}

	if (uid == 0 && ! reencrypt_only) {
		/*
		 * Root users store their key in /etc/$ROOTKEY so
		 * that they can auto reboot without having to be
		 * around to type a password.
		 */
		write_rootkey(secret);
	}
	if (! reencrypt_only)
		keylogin(netname, secret);
	exit(0);
}

void
usage()
{
	(void) fprintf(stderr,
		"usage: %s [-p] [-s nisplus | nis | files] \n",
		program_name);
	exit(1);
}


/*
 * Set the entry in the public key file
 */
static int
setpublicmap(char *netname,
	     char *public,
	     char *secret,
	     int database)
{
	char pkent[MAX_HEXKEYBYTES*2+2+KEYCHECKSUMSIZE];
	char *master;
	char *domain = NULL;
	nis_name nis_princ;

	(void) sprintf(pkent, "%s:%s", public, secret);
	switch (database) {
	case PK_YP:
		(void) yp_get_default_domain(&domain);
		if (yp_master(domain, PKMAP, &master) != 0) {
			(void) fprintf(stderr,
			"%s: cannot find master of NIS publickey database\n",
				program_name);
			exit(1);
		}
		(void) fprintf(stdout,
			"Sending key change request to %s ...\n", master);
		return (yp_update(domain, PKMAP, YPOP_STORE,
				netname, strlen(netname), pkent,
				strlen(pkent)));
	case PK_FILES:
		if (geteuid() != 0) {
			(void) fprintf(stderr,
		"%s: non-root users cannot change their key-pair in %s\n",
				program_name, PKFILE);
			exit(1);
		}
		return (localupdate(netname, PKFILE, YPOP_STORE, pkent));
	case PK_NISPLUS:
		nis_princ = get_nisplus_principal(nis_local_directory(),
				geteuid());
		return (nisplus_update(netname, public, secret, nis_princ));
	default:
		break;
	}
	return (1);
}

/*
 * populate 'newpass' and 'secret' and pass back
 * force will be only supported for a while
 * 	-- it is NOT documented as of s1093
 */
static int
getpasswd(uid_t uid, char *netname, int force, char *newpass, char *secret)
{
	char	*rpcpass = NULL, *x_rpcpass = NULL;
	char	*l_pass = NULL, *x_lpass = NULL;
	struct	passwd	*pw;
	struct	spwd	*spw;
	char	*login_passwd = NULL;
	char	prompt[256];
	char    *psec;


	/*
	 * get the shadow passwd from the repository
	 */
	if ((pw = getpwuid(uid)) == 0) {
		(void) fprintf(stderr,
			"%s: unable to locate passwd entry for uid %d\n",
			program_name, uid);
		return (0);
	}
	if ((spw = getspnam(pw->pw_name)) == 0) {
		(void) fprintf(stderr,
			"%s: unable to locate shadow passwd for uid %d\n",
			program_name, uid);
		return (0);
	}
	login_passwd = spw->sp_pwdp;

	/*
	 * now set effective uid to users uid
	 */
	seteuid(uid);

	/*
	 * get the old encryption passwd - this may or may not be
	 * their old login passwd
	 */
	(void) sprintf(prompt,
		"Please enter the Secure-RPC password for %s:",
		pw->pw_name);
	rpcpass = getpass(prompt);
	if (rpcpass == NULL) {
		(void) fprintf(stderr,
			"%s: key-pair unchanged for %s.\n",
			program_name, pw->pw_name);
		return (0);
	}

	/*
	 * get the secret key from the key respository
	 */
	if (! getsecretkey_long(netname, &psec, rpcpass)) {
		(void) fprintf(stderr,
			"%s: could not get secret key for '%s'\n",
			program_name, netname);
		return (0);
	}

	secret[0] = 0;
	if (psec)
		strcpy(secret, psec);
	free(psec);

	/*
	 * check if it is zero -> if it is then ask for the passwd again.
	 * then attempt to get the secret again
	 */
	if (secret[0] == 0) {
		(void) sprintf(prompt,
			"Try again. Enter the Secure-RPC password for %s:",
			pw->pw_name);
		rpcpass = getpass(prompt);
		if (rpcpass == NULL) {
			(void) fprintf(stderr,
				"%s: key-pair unchanged for %s.\n",
				program_name, pw->pw_name);
			return (0);
		}
		if (! getsecretkey_long(netname, &psec, rpcpass)) {
			(void) fprintf(stderr,
				"%s: could not get secret key for '%s'\n",
				program_name, netname);
			return (0);
		}
		if (!psec) {
			(void) fprintf(stderr,
				"%s: Unable to decrypt secret key for %s.\n",
				program_name, netname);
			return (0);
		}
		strcpy(secret, psec);
		free(psec);
	}
	/*
	 * check if 'user' has a key cached with keyserv.
	 * if there is no key cached, then keylogin the user.
	 * if uid == 0, might aswell write it to /etc/$ROOTKEY,
	 * assuming that if there is a /etc/$ROOTKEY then the
	 * roots' secret key should already be cached.
	 */
	if (! key_secretkey_is_set()) {
		keylogin(netname, secret);
		if ((uid == 0) && (reencrypt_only))
			write_rootkey(secret);
	}

	if (force) {
		/* simply get the new passwd - no checks */
		(void) sprintf(prompt, "Please enter New password:");
		l_pass = getpass(prompt);
		if (l_pass && (strlen(l_pass) != 0)) {
			strcpy(newpass, l_pass);
			return (1);
		}
		(void) fprintf(stderr,
			"%s: key-pair unchanged for %s.\n",
			program_name, pw->pw_name);
		return (0);
	}
	/*
	 * check if the secure-rpc passwd given above matches the
	 * the unix login passwd
	 */
	if (login_passwd && (strlen(login_passwd) != 0)) {
		/* NOTE: 1st 2 chars of an encrypted passwd = salt */
		x_rpcpass = crypt(rpcpass, login_passwd);
		if (strcmp(x_rpcpass, login_passwd) != 0) {
			/*
			 * the passwds don't match, get the unix
			 * login passwd
			 */
			(void) sprintf(prompt,
				"Please enter the login password for %s:",
				pw->pw_name);
			l_pass = getpass(prompt);
			if (l_pass && (strlen(l_pass) != 0)) {
				x_lpass = crypt(l_pass, login_passwd);
				if (strcmp(x_lpass, login_passwd) != 0) {
					/* try again ... */
					(void) sprintf(prompt,
			"Try again. Please enter the login password for %s:",
						pw->pw_name);
					l_pass = getpass(prompt);
					if (l_pass && (strlen(l_pass) != 0)) {
						x_lpass = crypt(l_pass,
							login_passwd);
						if (strcmp(x_lpass, l_pass)
								!= 0) {
							(void) fprintf(stderr,
								"Sorry.\n");
							return (0);
						}
						/* passwds match */
						strcpy(newpass, l_pass);
					} else {
						(void) fprintf(stderr,
					"%s: key-pair unchanged for %s.\n",
					program_name, pw->pw_name);
						return (0);
					}
				}
				/* passwds match */
				strcpy(newpass, l_pass);
			} else {
				/* need a passwd */
				(void) sprintf(prompt,
		"Need a password. Please enter the login password for %s:",
						pw->pw_name);
				l_pass = getpass(prompt);
				if (l_pass && (strlen(l_pass) != 0)) {
					x_lpass = crypt(l_pass, login_passwd);
					if (strcmp(x_lpass, l_pass) != 0) {
						(void) fprintf(stderr,
							"Sorry.\n");
						return (0);
					}
					/* passwds match */
					strcpy(newpass, l_pass);
				} else {
					(void) fprintf(stderr,
					"%s: key-pair unchanged for %s.\n",
					program_name, pw->pw_name);
					return (0);
				}
			}
		}
		/* rpc and login passwds match */
		strcpy(newpass, rpcpass);
	} else { 	/* no login passwd entry */
		(void) fprintf(stderr,
		"%s: no passwd found for %s in the shadow passwd entry.\n",
			program_name, pw->pw_name);
		return (0);
	}
	return (1);
}
