/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#pragma ident	"@(#)update.c	1.11	94/04/25 SMI"

/*
 * Copyright (C) 1986, Sun Microsystems, Inc.
 */

/*
 * Administrative tool to add a new user to the publickey database
 */
#include <stdio.h>
#include <unistd.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <rpcsvc/ypclnt.h>
#include <sys/wait.h>
#include <netdb.h>
#include <string.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <errno.h>
#include <malloc.h>

#define	MAXMAPNAMELEN 256

extern	char	program_name[];
extern	int	errno;

static	char	*basename(char *);
static int _openchild(char *command, FILE **fto, FILE **ffrom);
static int match(char *line, char *name);


static	char	SHELL[] = "/bin/sh";
#if 0
static	char	YPDBPATH[] = "/var/yp";
static	char	PKMAP[] = "publickey.byname";
static	char	PKFILE[] = "/etc/publickey";
#endif
static	char	UPDATEFILE[] = "updaters";
static	char	MAKE[] = "/usr/ccs/bin/make";

/*
 * Determine if requester is allowed to update the given map,
 * and update it if so. Returns the yp status, which is zero
 * if there is no access violation.
 */
int
mapupdate(char *name,
	char *mapname,
	u_int op,
	char *data)
{
	char	updater[MAXMAPNAMELEN + 40];
	FILE	*childargs;
	FILE	*childrslt;
#ifdef WEXITSTATUS
	int	status;
#else
	union wait status;
#endif
	pid_t	pid;
	u_int	yperrno;
	int	namelen, datalen;
	struct	stat	stbuf;

#ifdef DEBUG
	(void) fprintf(stderr, "%s %s\n", name, data);
#endif
	namelen = strlen(name);
	datalen = strlen(data);
	errno = 0;
	if (stat(MAKE, &stbuf) < 0)
		switch (errno) {
		case ENOENT:
			(void) fprintf(stderr,
			"%s: %s not found, please install on the system\n",
			program_name, MAKE);
			return (1);
		default:
			(void) fprintf(stderr,
				"%s: cannot access %s, errno=%d.\n",
				program_name, MAKE, errno);
			return (1);
		}
	(void) sprintf(updater, "%s -s -f %s %s",
			MAKE, UPDATEFILE, mapname);
	pid = _openchild(updater, &childargs, &childrslt);
	if (pid < 0)
		return (YPERR_YPERR);

	/*
	 * Write to child
	 */
	(void) fprintf(childargs, "%s\n", name);
	(void) fprintf(childargs, "%u\n", op);
	(void) fprintf(childargs, "%u\n", namelen);
	(void) fwrite(name, namelen, 1, childargs);
	(void) fprintf(childargs, "\n");
	(void) fprintf(childargs, "%u\n", datalen);
	(void) fwrite(data, datalen, 1, childargs);
	(void) fprintf(childargs, "\n");
	(void) fclose(childargs);

	/*
	 * Read from child
	 */
	(void) fscanf(childrslt, "%d", &yperrno);
	(void) fclose(childrslt);

	(void) wait(&status);
#ifdef WEXITSTATUS
	if (WEXITSTATUS(status) != 0) {
#else
	if (status.w_retcode != 0) {
#endif
		return (YPERR_YPERR);
	}
	return (yperrno);
}

/*
 * returns pid, or -1 for failure
 */
static int
_openchild(char *command,
	FILE **fto,
	FILE **ffrom)
{
	int i;
	pid_t pid;
	int pdto[2];
	int pdfrom[2];
	char *com;
	struct rlimit rl;

	if (pipe(pdto) < 0) {
		goto error1;
	}
	if (pipe(pdfrom) < 0) {
		goto error2;
	}
#ifdef VFORK
	switch (pid = vfork()) {
#else
	switch (pid = fork()) {
#endif
	case -1:
		goto error3;

	case 0:
		/*
		 * child: read from pdto[0], write into pdfrom[1]
		 */
		(void) close(0);
		(void) dup(pdto[0]);
		(void) close(1);
		(void) dup(pdfrom[1]);
		getrlimit(RLIMIT_NOFILE, &rl);
		for (i = rl.rlim_max - 1; i >= 3; i--) {
			(void) close(i);
		}
		com = malloc((unsigned) strlen(command) + 6);
		if (com == NULL) {
			_exit(~0);
		}
		(void) sprintf(com, "exec %s", command);
		execl(SHELL, basename(SHELL), "-c", com, NULL);
		_exit(~0);

	default:
		/*
		 * parent: write into pdto[1], read from pdfrom[0]
		 */
		*fto = fdopen(pdto[1], "w");
		(void) close(pdto[0]);
		*ffrom = fdopen(pdfrom[0], "r");
		(void) close(pdfrom[1]);
		break;
	}
	return (pid);

	/*
	 * error cleanup and return
	 */
error3:
	(void) close(pdfrom[0]);
	(void) close(pdfrom[1]);
error2:
	(void) close(pdto[0]);
	(void) close(pdto[1]);
error1:
	return (-1);
}

static char *
basename(char *path)
{
	char	*p;

	p = strrchr(path, '/');
	if (p == NULL)
		return (path);
	return (p + 1);
}

/*
 * Determine if requester is allowed to update the given map,
 * and update it if so. Returns the status, which is zero
 * if there is no access violation, 1 otherwise.
 * This function updates the local file.
 */
int
localupdate(char *name,	/* user the entry belongs to */
	char *filename,
	u_int op,
	char *data)
{
	char	line[MAX_HEXKEYBYTES*2+KEYCHECKSUMSIZE+MAXNETNAMELEN+128];
	FILE	*rf;
	FILE	*wf;
	char	tmpname[80];
	int	err;

	/*
	 * Check permission
	 */
	if (strcmp(name, "nobody") == 0) {
		/* cannot change keys for nobody */
		(void) fprintf(stderr,
			"%s: cannot change key-pair for %s\n",
			program_name, name);
		return (1);
	}

	/*
	 * Open files
	 */
	(void) memset(tmpname, 0, 80);
	(void) sprintf(tmpname, "%s.tmp", filename);
	rf = fopen(filename, "r");
	if (rf == NULL) {
		(void) fprintf(stderr,
		"%s: cannot read %s\n", program_name, filename);
		return (1);
	}
	wf = fopen(tmpname, "w");
	if (wf == NULL) {
		(void) fprintf(stderr, "%s: cannot create '%s' to write to.\n",
			program_name, tmpname);
		return (1);
	}
	err = -1;
	while (fgets(line, sizeof (line), rf)) {
		if (err < 0 && match(line, name)) {
			switch (op) {
			case YPOP_INSERT:
				err = 1;
				break;
			case YPOP_STORE:
			case YPOP_CHANGE:
				(void) fprintf(wf, "%s\t%s\n", name, data);
				err = 0;
				break;
			case YPOP_DELETE:
				/* do nothing */
				err = 0;
				break;
			}
		} else {
			fputs(line, wf);
		}
	}
	if (err < 0) {
		switch (op) {
		case YPOP_CHANGE:
		case YPOP_DELETE:
			err = 1;
			break;
		case YPOP_INSERT:
		case YPOP_STORE:
			err = 0;
			(void) fprintf(wf, "%s\t%s\n", name, data);
			break;
		}
	}
	fclose(wf);
	fclose(rf);
	if (err == 0) {
		if (rename(tmpname, filename) < 0) {
			(void) fprintf(stderr,
				"%s: cannot rename %s to %s\n",
				program_name, tmpname, filename);
			return (1);
		}
	} else {
		if (unlink(tmpname) < 0) {
			(void) fprintf(stderr,
				"%s: cannot delete %s\n",
				program_name, tmpname);
			return (1);
		}
	}
	return (err);
}

static int
match(char *line, char *name)
{
	int	len;

	len = strlen(name);
	return (strncmp(line, name, len) == 0 &&
		(line[len] == ' ' || line[len] == '\t'));
}
