/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#ident	"@(#)publickey.c	1.18	94/08/23 SMI"  /*  SVr4 1.2 */

/*
 * Copyright (c) 1986-1992 by Sun Microsystems Inc.
 */

/*
 * publickey.c
 *
 *
 * Public and Private (secret) key lookup routines. These functions
 * are used by the secure RPC auth_des flavor to get the public and
 * private keys for secure RPC principals. Originally designed to
 * talk only to YP, AT&T modified them to talk to files, and now
 * they can also talk to NIS+. The policy for these lookups is now
 * defined in terms of the nameservice switch as follows :
 *	publickey: nis files
 *
 * Note :
 * 1.  NIS+ combines the netid.byname and publickey.byname maps
 *	into a single NIS+ table named cred.org_dir
 * 2.  To use NIS+, the publickey needs to be
 *	publickey: nisplus
 *	(or have nisplus as the first entry).
 *	The nsswitch.conf file should be updated after running nisinit
 *	to reflect this.
 */
#include "../rpc/rpc_mt.h"
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <rpc/trace.h>
#include <pwd.h>
#include "nsswitch.h"
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <rpcsvc/nis.h>
#include <rpcsvc/ypclnt.h>
#include <malloc.h>

#include <rpc/keylib.h>

static const char* PKTABLE = "cred.org_dir";
static const char* PKMAP = "publickey.byname";
static const char* PKFILE = "/etc/publickey";
#define	PKTABLE_LEN 12
#ifndef NO_ANK_FIX
#define	WORKBUFSIZE (2*MAX_HEXKEYBYTES + KEYCHECKSUMSIZE + MAXNETNAMELEN + 128)
#else
#define	WORKBUFSIZE 1024
#endif

/* default publickey policy:
 *	publickey: nis [NOTFOUND = return] files
 */


/*	NSW_NOTSUCCESS  NSW_NOTFOUND   NSW_UNAVAIL    NSW_TRYAGAIN */
#define	DEF_ACTION {__NSW_RETURN, __NSW_RETURN, __NSW_CONTINUE, __NSW_CONTINUE}

static struct __nsw_lookup lookup_files = {"files", DEF_ACTION, NULL, NULL},
		lookup_nis = {"nis", DEF_ACTION, NULL, &lookup_files};
static struct __nsw_switchconfig publickey_default =
			{0, "publickey", 2, &lookup_nis};

#ifndef NUL
#define	NUL '\0'
#endif

extern mutex_t serialize_pkey;

/*
 * extract_secret()
 *
 * This generic function will extract the private key
 * from a string using the given password. Note that
 * it uses the DES based function xdecrypt()
 */
static int
extract_secret(
	char		*raw,
	keybuf_long	*private,
	char		*passwd
#ifdef SECURE_PUBLICKEY
	, char		*name
#endif
	       )
{
	char	buf[WORKBUFSIZE]; /* private buffer to work from */
	char	*p;
	int     klen;

	trace1(TR_extract_secret, 0);
	if (! passwd || ! raw || ! private) {
		if (private)
			*private = NUL;
		trace1(TR_extract_secret, 1);
		return (0);
	}

	strcpy(buf, raw);
	p = strchr(buf, '\n');
	if (p)
		*p = 0;
	p = strchr(buf, ':');
	if (p) {
		*p = 0;
	}
	klen = strlen(buf) - KEYCHECKSUMSIZE;
	if (klen < 0)
		return 0;
#ifndef SECURE_PUBLICKEY
	if (!xdecrypt(buf, passwd)) {
#else
	if (!xxdecrypt(buf, passwd, name)) {
#endif
		*private = NULL;
		trace1(TR_extract_secret, 1);
		return (1);
	}
	
	if (memcmp(buf, &buf[klen], KEYCHECKSUMSIZE) != 0) {
		*private = NULL;
		trace1(TR_extract_secret, 1);
		return (1);
	}

	buf[klen] = 0;
	*private = strdup(buf);
	trace1(TR_extract_secret, 1);
	return (1);
}

/*
 * These functions are the "backends" for the switch for public keys. They
 * get both the public and private keys from each of the supported name
 * services (nis, nisplus, files). They are passed the appropriate parameters
 * and return 0 if unsuccessful with *errp set, or 1 when they got just the
 * public key and 3 when they got both the public and private keys.
 *
 *
 * getkey_nis()
 *
 * Internal implementation of getpublickey() using NIS (aka Yellow Pages,
 * aka YP).
 *
 * NOTE : *** this function returns nsswitch codes and _not_ the
 * value returned by getpublickey.
 */
static int
getkeys_nis_long(
	int  *errp,
	netnamestr netname,
	keybuf_long *pkey,
	keybuf_long *skey,
	char *passwd)
{
	char 	*domain;
	char	*keyval = NULL;
	int	keylen, err, r = 0;
	char	*p;

	trace1(TR_getkeys_nis, 0);

	p = strchr(netname, '@');
	if (! p) {
		*errp = __NSW_UNAVAIL;
		trace1(TR_getkeys_nis, 1);
		return (0);
	}

	domain = ++p;
	err = yp_match(domain, (char *)PKMAP, netname, strlen(netname),
			&keyval, &keylen);
	switch (err) {
	case YPERR_KEY :
		if (keyval)
			free(keyval);
		*errp = __NSW_NOTFOUND;
		trace1(TR_getkeys_nis, 1);
		return (0);
	default :
		if (keyval)
			free(keyval);
		*errp = __NSW_UNAVAIL;
		trace1(TR_getkeys_nis, 1);
		return (0);
	case 0:
		break;
	}

	p = strchr(keyval, ':');
	if (p == NULL) {
		free(keyval);
		*errp = __NSW_NOTFOUND;
		trace1(TR_getkeys_nis, 1);
		return (0);
	}
	*p = 0;
	if (pkey)
		*pkey = strdup(keyval);
	r = 1;
	p++;
#ifndef SECURE_PUBLICKEY
	if (skey && extract_secret(p, skey, passwd))
#else
	if (skey && extract_secret(p, skey, passwd, netname))
#endif
		r |= 2;
	free(keyval);
	*errp = __NSW_SUCCESS;
	trace1(TR_getkeys_nis, 1);
	return (r);
}

/*
 * getkey_files()
 *
 * The files version of getpublickey. This function attempts to
 * get the publickey from the file PKFILE .
 *
 * This function defines the format of the /etc/publickey file to
 * be :
 *	netname <whitespace> publickey:privatekey
 *
 * NOTE : *** this function returns nsswitch codes and _not_ the
 * value returned by getpublickey.
 */

static int
getkeys_files_long(
	int	*errp,
	netnamestr	netname,
	keybuf_long	*pkey,
	keybuf_long	*skey,
	char	*passwd)
{
	register char *mkey;
	register char *mval;
	char buf[WORKBUFSIZE];
	int	r = 0;
	char *res;
	FILE *fd;
	char *p;

	trace1(TR_getkeys_files, 0);

	fd = fopen(PKFILE, "r");
	if (fd == (FILE *) 0) {
		*errp = __NSW_UNAVAIL;
		trace1(TR_getkeys_files, 1);
		return (0);
	}

	/* Search through the file linearly :-( */
	while ((res = fgets(buf, WORKBUFSIZE, fd)) != NULL) {

		if ((res[0] == '#') || (res[0] == '\n'))
			continue;
		else {
			mkey = strtok(buf, "\t ");
			if (mkey == NULL) {
				syslog(LOG_INFO,
				"getpublickey: Bad record in %s for %s",
							PKFILE, netname);
				continue;
			}
			mval = strtok((char *)NULL, " \t#\n");
			if (mval == NULL) {
				syslog(LOG_INFO,
				"getpublickey: Bad record in %s for %s",
							PKFILE, netname);
				continue;
			}
			/* NOTE : Case insensitive compare. */
			if (strcasecmp(mkey, netname) == 0) {
				p = strchr(mval, ':');
				if (p == NULL) {
					syslog(LOG_INFO,
				"getpublickey: Bad record in %s for %s",
							PKFILE, netname);
					continue;
				}

				*p = 0;
				if (pkey)
					*pkey = strdup(mval);
				r = 1;
				p++;
#ifndef SECURE_PUBLICKEY
				if (skey && extract_secret(p, skey, passwd))
#else
				if (skey && extract_secret(p, skey, passwd, netname))
#endif
					r |= 2;
				fclose(fd);
				*errp = __NSW_SUCCESS;
				trace1(TR_getkeys_files, 1);
				return (r);
			}
		}
	}

	fclose(fd);
	*errp = __NSW_NOTFOUND;
	trace1(TR_getkeys_files, 1);
	return (0);
}

/*
 * getkeys_nisplus()
 *
 * This one fetches the key pair from NIS+.
 */
static int
getkeys_nisplus_long(
	int 	*err,
	netnamestr	netname,
	keybuf_long	*pkey,
	keybuf_long	*skey,
	char	*passwd)
{
	nis_result	*res;
	int		r = 0;
	char		*domain, *p;
	char		buf[NIS_MAXNAMELEN+1];
	int len;

	trace1(TR_getkeys_nisplus, 0);

	domain = strchr(netname, '@');
	if (! domain) {
		*err = __NSW_UNAVAIL;
		trace1(TR_getkeys_nisplus, 1);
		return (0);
	}
	domain++;

	if ((strlen(netname)+PKTABLE_LEN+strlen(domain)+32) >
		(size_t) NIS_MAXNAMELEN) {
		*err = __NSW_UNAVAIL;
		trace1(TR_getkeys_nisplus, 1);
		return (0);
	}

	/*
	 * Cred table has following format for DES entries:
	 * cname   auth_type auth_name public  private
	 * ----------------------------------------------------------
	 * nisname	DES	netname	pubkey	prikey
	 */
	sprintf(buf, "[auth_name=\"%s\",auth_type=DES],%s.%s",
		netname, PKTABLE, domain);
	if (buf[strlen(buf)-1] != '.')
	strcat(buf, ".");

	/*
	 * If we want a publickey, then we need the connection to be
	 * authenticated; otherwise, since the secretkey is encrypted
	 * it can be gotten from an unauthenticated connection.
	 * Furthermore, the latter needs to be the case if
	 * the bootstrapping process is to work.  More specifically,
	 * when you're doing a keylogin, you need to get your secretkey
	 * and you *cannot* use an authenticated handle to get it because
	 * you don't have your secretkey yet.
	 */
	if (pkey)
		res = nis_list(buf,
				USE_DGRAM+NO_AUTHINFO+FOLLOW_LINKS+FOLLOW_PATH,
				NULL, NULL);
	else
		res = nis_list(buf,
				USE_DGRAM+NO_AUTHINFO+FOLLOW_LINKS+FOLLOW_PATH,
				NULL, NULL);
	switch (res->status) {
	case NIS_SUCCESS:
	case NIS_S_SUCCESS:
		break;
	case NIS_NOTFOUND:
	case NIS_PARTIAL:
	case NIS_NOSUCHNAME:
	case NIS_NOSUCHTABLE:
		nis_freeresult(res);
		*err = __NSW_NOTFOUND;
		trace1(TR_getkeys_nisplus, 1);
		return (0);
	case NIS_S_NOTFOUND:
	case NIS_TRYAGAIN:
		syslog(LOG_ERR, "getkeys: (nis+ key lookup): %s\n",
			nis_sperrno(res->status));
		nis_freeresult(res);
		*err = __NSW_TRYAGAIN;
		trace1(TR_getkeys_nisplus, 1);
		return (0);
	default:
		*err = __NSW_UNAVAIL;
		syslog(LOG_ERR,
			"getkeys: (nis+ key lookup): %s\n",
			nis_sperrno(res->status));
		nis_freeresult(res);
		trace1(TR_getkeys_nisplus, 1);
		return (0);
	}

	if (pkey) {
		len = ENTRY_LEN(res->objects.objects_val, 3);
		*pkey = (char*)malloc(len + 1);
		strncpy(*pkey, ENTRY_VAL(res->objects.objects_val, 3), len);
		(*pkey)[len] = NUL;

		/*
		 * XXX
		 * This is only for backward compatibility with the old cred
		 * table format.  The new one does not have a ':'.
		 */
		p = strchr(*pkey, ':');
		if (p)
			*p = NUL;
	}
	r = 1; /* At least public key was found; always true at this point */

#ifndef SECURE_PUBLICKEY
	if (skey && extract_secret(ENTRY_VAL(res->objects.objects_val, 4),
				skey, passwd))
#else
	if (skey && extract_secret(ENTRY_VAL(res->objects.objects_val, 4),
				skey, passwd, netname))
#endif
		r |= 2;

	nis_freeresult(res);
	*err = __NSW_SUCCESS;
	trace1(TR_getkeys_nisplus, 1);
	return (r);
}

/*
 * getpublickey(netname, key)
 *
 * This is the actual exported interface for this function.
 */

int
getpublickey_long(netnamestr netname, keybuf_long *pkey)
{
	int	needfree = 1, res, err;
	struct __nsw_switchconfig *conf;
	struct __nsw_lookup *look;
	enum __nsw_parse_err perr;

	trace1(TR_getpublickey, 0);
	if (! netname || ! pkey) {
		trace1(TR_getpublickey, 1);
		return (0);
	}

	mutex_lock(&serialize_pkey);

	conf = __nsw_getconfig("publickey", &perr);
	if (! conf) {
		conf = &publickey_default;
		needfree = 0;
	}
	for (look = conf->lookups; look; look = look->next) {
		if (strcmp(look->service_name, "nisplus") == 0)
			res = getkeys_nisplus_long(&err, netname, pkey, NULL, NULL);
		else if (strcmp(look->service_name, "nis") == 0)
			res = getkeys_nis_long(&err, netname, pkey, NULL, NULL);
		else if (strcmp(look->service_name, "files") == 0)
			res = getkeys_files_long(&err, netname, pkey, NULL, NULL);
		else {
			syslog(LOG_INFO, "Unknown publickey nameservice '%s'",
						look->service_name);
			err = __NSW_UNAVAIL;
			res = 0;
		}
		switch (look->actions[err]) {
		case __NSW_CONTINUE :
			continue;
		case __NSW_RETURN :
			if (needfree)
				__nsw_freeconfig(conf);
			mutex_unlock(&serialize_pkey);
			trace1(TR_getpublickey, 1);
			return ((res & 1) != 0);
		default :
			syslog(LOG_INFO, "Unknown action for nameservice %s",
					look->service_name);
		}
	}

	if (needfree)
		__nsw_freeconfig(conf);
	mutex_unlock(&serialize_pkey);
	trace1(TR_getpublickey, 1);
	return (0);
}

int
getsecretkey_long(netnamestr netname, keybuf_long *skey, char *passwd)
{
	int	needfree = 1, res, err;
	struct __nsw_switchconfig *conf;
	struct __nsw_lookup *look;
	enum __nsw_parse_err perr;

	trace1(TR_getsecretkey, 0);
	if (! netname || !skey) {
		trace1(TR_getsecretkey, 1);
		return (0);
	}

	mutex_lock(&serialize_pkey);

	conf = __nsw_getconfig("publickey", &perr);

	if (! conf) {
		conf = &publickey_default;
		needfree = 0;
	}

	for (look = conf->lookups; look; look = look->next) {
		if (strcmp(look->service_name, "nisplus") == 0)
			res = getkeys_nisplus_long(&err, netname, NULL,
								skey, passwd);
		else if (strcmp(look->service_name, "nis") == 0)
			res = getkeys_nis_long(&err, netname, NULL, skey, passwd);
		else if (strcmp(look->service_name, "files") == 0)
			res = getkeys_files_long(&err, netname, NULL, skey, passwd);
		else {
			syslog(LOG_INFO, "Unknown publickey nameservice '%s'",
						look->service_name);
			err = __NSW_UNAVAIL;
			res = 0;
		}
		switch (look->actions[err]) {
		case __NSW_CONTINUE :
			continue;
		case __NSW_RETURN :
			if (needfree)
				__nsw_freeconfig(conf);
			mutex_unlock(&serialize_pkey);
			trace1(TR_getsecretkey, 1);
			return ((res & 2) != 0);
		default :
			syslog(LOG_INFO, "Unknown action for nameservice %s",
					look->service_name);
		}
	}
	if (needfree)
		__nsw_freeconfig(conf);
	mutex_unlock(&serialize_pkey);
	trace1(TR_getsecretkey, 1);
	return (0);
}

#if 0
int
getsecretkey(char * netname, char * secretkey, char * passwd)
{
	char * skey;
	if (!getsecretkey_long(netname, &skey, passwd))
		return 0;
	if (!skey) {
		secretkey[0] = 0;
		return 1;
	}
	if (strlen(skey) != DEFAULT_HEXKEYBYTES) {
		free(skey);
		secretkey[0] = 0;
		return 1;
	}
	strcpy(secretkey, skey);
	free(skey);
	return 0;
}

int
getpublickey(char * netname, char * publickey)
{
	char * pkey;

	if (!getpublickey_long(netname, &pkey))
		return 0;
	if (!pkey)
		return 0;
	strncpy(publickey, pkey, DEFAULT_HEXKEYBYTES);
	publickey[DEFAULT_HEXKEYBYTES] = 0;
	free(pkey);
	return 0;
}
#endif
